unit Main;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ComCtrls, System.Math;

type
  TMainFrm = class(TForm)
    GenBtn: TButton;
    ViewBtn: TButton;
    OpenDlg: TOpenDialog;
    ViewOpenDlg: TOpenDialog;
    procedure FormCreate(Sender: TObject);
    procedure GenBtnClick(Sender: TObject);
    procedure ViewBtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

  TTower = record
    X, Y: Integer;
    Typ: String;
  end;

  TSides = record
    Xp, Yp, Xn, Yn: Boolean;
  end;

  TBlock = record
    Color: TColor;
    Side: TSides;
  end;

  TBuildBlock = record
    Color: Char;
    dX, dY: Integer;
  end;

  TBlocks = Array[1..2, 1..4, 1..5] of Integer;

var
  MainFrm: TMainFrm;

implementation

uses View;

{$R *.dfm}

function IndexColor(I: Integer): TColor;
begin
  case I of
    1: Result := clAqua;
    2: Result := clRed;
    3: Result := clGreen;
    4: Result := clBlue;
    5: Result := clYellow;
  else Result := clBlack;
  end;
end;

function ColorIndex(Ch: Char): Integer;
begin
  case Ch of
    'a', 'A': Result := 1;
    'r', 'R': Result := 2;
    'g', 'G': Result := 3;
    'b', 'B': Result := 4;
    'y', 'Y': Result := 5;
  else Result := 0;
  end;
end;

function ColorChar(I: Integer): Char;
begin
  case I of
    1: Result := 'A';
    2: Result := 'R';
    3: Result := 'G';
    4: Result := 'B';
    5: Result := 'Y';
  else Result := 'X';
  end;
end;

function InArray(Ch: Char; A: Array of char): Boolean;
var
  i: Integer;
begin
  for i := 0 to High(A) do begin
    if (A[i] = Ch) then begin
      Result := True;
      Exit;
    end;
  end;
  Result := False;
end;

procedure StringSplit(const Separator: Array of char; S: String; var Values: TArray<String>);
var
  i: Integer;
  Temp: String;

begin
  SetLength(Values, 0);
  Temp := '';

  for i := 1 to Length(S)+1 do begin
    if (i = Length(S)+1) then begin
      SetLength(Values, Length(Values) + 1);
      Values[High(Values)] := Temp;
      Temp := '';
    end else if (InArray(S[i], Separator)) then begin
      SetLength(Values, Length(Values) + 1);
      Values[High(Values)] := Temp;
      Temp := '';
    end else begin
      Temp := Temp + S[i];
    end;
  end;
end;

procedure TMainFrm.FormCreate(Sender: TObject);
begin
  OpenDlg.InitialDir := ExtractFileDir(Application.ExeName);
  ViewOpenDlg.InitialDir := ExtractFileDir(Application.ExeName);
end;

function ReadValues(Line: String): TArray<String>; inline;
var
  Values: TArray<String>;
begin
  Line := Trim(Line);

  while (Pos('  ', Line) > 0) do begin
    Line := StringReplace(Line, '  ', ' ', [rfReplaceAll]);
  end;

  StringSplit([' '], Line, Values);

  Result := Values;
end;

procedure BuildTower(FileName: String; TowerInfo: TTower; Blocks: TBlocks);
var
  Output: TextFile;
  Tower: Array of Array of Array of TBuildBlock;
  TowerBin: Array of Array of Array of Boolean;
  i, j, k, c, prevk: Integer;
  Cont: Boolean;
begin

  try
    AssignFile(Output, Copy(FileName, 1, Length(FileName) - Length(ExtractFileExt(FileName)))+'.20');
    ReWrite(Output);

    Writeln(Output, IntToStr(TowerInfo.X)+' '+IntToStr(TowerInfo.Y)+' '+TowerInfo.Typ);

    SetLength(Tower, TowerInfo.X);
    SetLength(TowerBin, TowerInfo.X);

    for i := 0 to High(Tower) do begin
      SetLength(Tower[i], TowerInfo.Y);
      SetLength(TowerBin[i], TowerInfo.Y);
    end;

    repeat
      Cont := True;
      for i := 0 to High(Tower) do begin
        for j := 0 to High(Tower[i]) do begin
          if ((LowerCase(TowerInfo.Typ[1]) = 'd') AND (i > 0) AND (i < High(Tower))) AND (j > 0) AND (j < High(Tower[i])) then break;
          for k := 1 to 5 do begin
            c := k;
            if ((Length(TowerInfo.Typ) > 1) AND (LowerCase(TowerInfo.Typ[2]) = 'b') AND (c = prevk)) then begin
              repeat
                c := c + 1;
                if (c > 5) then c := 1;
              until (c = prevk) OR (Blocks[1][1][c] > 0);
            end;
            if (Blocks[1][1][c] > 0) then begin
              SetLength(Tower[i][j], Length(Tower[i][j]) + 1);
              With Tower[i][j][High(Tower[i][j])] do begin
                Color := ColorChar(c);
              end;
              Writeln(Output, IntToStr(High(Tower[i][j])+1)+' '+IntToStr(i+1)+' '+IntToStr(j+1)+' '+Tower[i][j][High(Tower[i][j])].Color+' 1x1');

              Blocks[1][1][c] := Blocks[1][1][c] - 1;

              Cont := True;
              prevk := c;

              Break;
            end else begin
              Cont := False;
            end;
          end;
        end;
      end;
    until NOT(Cont);
  finally
    CloseFile(Output);
  end;
end;

procedure TMainFrm.GenBtnClick(Sender: TObject);
var
  Progress: TProgressBar;
  Files: TStringList;
  Search: TSearchRec;
  Spec: TextFile;
  i: Integer;
  Line: String;
  Values, Dims: TArray<String>;
  Tower: TTower;
  Blocks: TBlocks;
  X, Y: Integer;
begin
  if (OpenDlg.Execute) then begin
    Progress := TProgressBar.Create(Self);
    With Progress do begin
      Parent := Self;
      Top := Self.Height;
      Left := 8;
      Height := 17;
      Width := Self.Width - 16;

      Position := 0;
    end;

    Application.ProcessMessages;

    Files := TStringList.Create;

    try
      Self.Height := Self.Height + 17 + 8;

      GenBtn.Enabled := False;
      ViewBtn.Enabled := False;

      //////////////
      if (FindFirst(ExtractFilePath(OpenDlg.FileName)+'*', faAnyFile, Search) = 0) then begin
        if (LowerCase(ExtractFileExt(Search.Name)) = '.in') then Files.Add(Search.Name);

        while (FindNext(Search) = 0) do begin
          if (LowerCase(ExtractFileExt(Search.Name)) = '.in') then Files.Add(Search.Name);
        end;
      end;

      Progress.Max := Files.Count;

      Files.Sort;

      for i := 0 to Files.Count - 1 do begin
        try
          AssignFile(Spec, ExtractFilePath(OpenDlg.FileName)+Files[i]);
          Reset(Spec);

          Readln(Spec, Line);

          Values := ReadValues(Line);

          Tower.X := StrToInt(Values[0]);
          Tower.Y := StrToInt(Values[1]);
          Tower.Typ := Values[2];

          while NOT(Eof(Spec)) do begin
            Readln(Spec, Line);

            Values := ReadValues(Line);

            StringSplit(['x', 'X'], Values[1], Dims);

            X := StrToInt(Dims[0]);
            Y := StrToInt(Dims[1]);

            Blocks[Min(X, Y), Max(X, Y), ColorIndex(Values[0][1])] := StrToInt(Values[2]);
          end;

          BuildTower(ExtractFilePath(OpenDlg.FileName)+Files[i], Tower, Blocks);
        finally
          CloseFile(Spec);
        end;

        Progress.Position := Progress.Position + 1;
        Application.ProcessMessages;
      end;
    finally
      Progress.Free;
      Self.Height := Self.Height - 17 - 8;

      GenBtn.Enabled := True;
      ViewBtn.Enabled := True;

      Files.Free;
    end;
  end;
end;

procedure TMainFrm.ViewBtnClick(Sender: TObject);
var
  Plan: TextFile;
  Values, Dims: TArray<String>;
  Line: String;
  i, j, dX, dY, X, Y, Z: Integer;
begin
  if (ViewOpenDlg.Execute) then begin
    try
      AssignFile(Plan, ViewOpenDlg.FileName);
      Reset(Plan);

      Readln(Plan, Line);
      Values := ReadValues(Line);

      SetLength(ViewFrm.Tower, StrToInt(Values[0]));
      for i := 0 to High(ViewFrm.Tower) do begin
        SetLength(ViewFrm.Tower[i], StrToInt(Values[1]));
      end;

      while NOT(Eof(Plan)) do begin
        Readln(Plan, Line);
        Values := ReadValues(Line);

        X := StrToInt(Values[1]);
        Y := StrToInt(Values[2]);
        Z := StrToInt(Values[0]);

        StringSplit(['x', 'X'], Values[4], Dims);
        dX := StrToInt(Dims[0]);
        dY := StrToInt(Dims[1]);

        for i := 1 to dX do begin
          for j := 1 to dY do begin
            if (Length(ViewFrm.Tower[X+i-1][Y+j-1]) < Z) then
              SetLength(ViewFrm.Tower[X+i-1][Y+j-1], Z);

            With ViewFrm.Tower[X+i-1][Y+j-1][Z] do begin
              Color := IndexColor(ColorIndex(Values[3][1]));
              With Side do begin
                Xp := NOT((dX > 1) AND (i < dX));
                Yp := NOT((dY > 1) AND (j < dY));
                Xn := NOT(i > 1);
                Yn := NOT(j > 1);
              end;
            end;
          end;
        end;
      end;
    finally
      CloseFile(Plan);

      ViewFrm.ShowModal;
      ViewFrm.Draw(0);
    end;
  end;
end;

end.
